"""Elegant astronomy for Python

Most users will use Skyfield by importing ``skyfield.api`` and using the
functions and classes there.  See the ``documentation`` directory inside
the source code, as well as the http://rhodesmill.org/skyfield/ site!

"""
VERSION = (0, 9)
__version__ = '.'.join(map(str, VERSION))
